#!/bin/bash
#  
#  For more information, please see: http://software.sci.utah.edu
#  
#  The MIT License
#  
#  Copyright (c) 2009 Scientific Computing and Imaging Institute,
#  University of Utah.
#  
#  License for the specific language governing rights and limitations under
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#  
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#  
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#  
#    File   : build.sh
#    Author : McKay Davis
#    Date   : Tue Dec 19 15:35:50 2006


# SCIRun build script
#
# This script will build SCIRun from scratch
#

INSIGHT_TOOLKIT=InsightToolkit-3.10.1
WXWIDGETS_TOOLKIT_MAC=wxMac-2.8.9
WXWIDGETS_TOOLKIT_ARG_MAC="--with-mac"
WXWIDGETS_TOOLKIT_GTK=wxGTK-2.8.9
WXWIDGETS_TOOLKIT_ARG_GTK="--with-gtk"
CMAKE=cmake-2.6.2

printhelp() {
    echo -e "--cmake-args=[string]\tCommand line arguments to CMake"
    echo -e "--debug\t\t\tBuilds SCIRun and ITK with debug symbols"
    echo -e "--release\t\tBuilds SCIRun and ITK without debug symbols (default)"
    echo -e "--with-itk\t\tDownloads and builds $INSIGHT_TOOLKIT and the Insight SCIRun package"
    echo -e "--with-wxwidgets\tDownloads and builds wxWidgets (used by Seg3D)"
    echo -e "--without-meshing-pipeline\tDownloads necessary thirdparties and builds meshing pipeline"
    echo -e "--without-tetgen\tTurns off tetgen support. By default, tetgen is built with SCIRun"
    echo -e "--with-lapack\t\tTurn on SCIRun LAPACK support."
    echo -e "--get-cmake\t\tDownloads and builds $CMAKE, regarless of the version installed on the system"
    echo -e "--static\t\tBuilds static libs"
    echo -e "--seg3d-only\t\tOnly builds static version of Seg3D PowerApp"
    echo -e "--packages=[string]\tComma seperated string specifies which SCIRun packages are built"
    echo -e "--no-ctest\t\tDo not run ctest"
    echo -e "--64bit\t\t\tForce a 64bit SCIRun version [OSX only]"
    echo -e "--universal\t\tEnable universal build for Mac OSX"
    echo -e "--documentation\t\tEnable building documentation (requires LaTeX)"
    echo -e "-j#\t\t\tRuns # parallel make processes when building"
    echo -e "-?\t\t\tThis help"
    exit 0
}

# will cause the script to bailout if the passed in command fails
try () {
  $*
  if [ $? != "0" ]; then
      echo -e "\n***ERROR in build script\nThe failed command was:\n$*\n"
      exit 1
  fi
}

trybuild () {
  $*
  if [ $? != "0" ]; then
      echo -e "Building SCIRun returned an error\n"
      echo -e "Either the code failed to build properly or\n"
      echo -e "the testing programs failed to complete without\n"
      echo -e "every single test program passing the test.\n"
      exit 1
  fi
}
# functionally equivalent to try(),
# but it prints a different error message
ensure () {
  $* >& /dev/null
  if [ $? != "0" ]; then
      echo -e "\n***ERROR, $* is required but not found on this system\n"
      exit 1
  fi
}

# Try to find a version of cmake
find_cmake() {
    if [ "$getcmake" != "1" ]; then
        cmakebin=`which cmake`
        ctestbin=`which ctest`
    fi
 
    #if it is not found
    if [ ! -e "$cmakebin" ]; then
    # then look for our own copy made by this script previously
        cmakebin=$DIR/cmake/local/bin/cmake
        ctestbin=$DIR/cmake/local/bin/ctest
        try mkdir -p $DIR/cmake/
        try cd $DIR/cmake
        if [ ! -e "$cmakebin" ]; then
        # try to download and build our own copy in local
            try $getcommand http://www.cmake.org/files/v2.6/${CMAKE}.tar.gz
            try tar xvzf ${CMAKE}.tar.gz
            try cd $CMAKE
            try ./configure --prefix=$DIR/cmake/local
            try make $makeflags
            try make install
        fi
    fi
    
    ctestbin="$ctestbin --ctest-config=${DIR}/src/CTestConfig.cmake"
    
    echo cmakebin=$cmakebin
    echo ctestbin=$ctestbin
    ensure $cmakebin --version
}

build_ITK() {
    if [ $builditk == "1" ]; then
        try cd $DIR
        if [ ! -e "$DIR/${INSIGHT_TOOLKIT}.tar.gz" ]; then 
            try echo "Downloading the Insight Toolkit (ITK)..."
            try $getcommand http://downloads.sourceforge.net/itk/${INSIGHT_TOOLKIT}.tar.gz
        fi
        
        if [ ! -e "$DIR/$INSIGHT_TOOLKIT" ]; then
            try tar xvzf ${INSIGHT_TOOLKIT}.tar.gz
        fi
        
        if [ "$osx" == "0" ]; then
          try cd $DIR/$INSIGHT_TOOLKIT
          try $cmakebin . -DBUILD_EXAMPLES=0 -DBUILD_TESTING=0 -DCMAKE_INSTALL_PREFIX=$SCIRUN_THIRDPARTY_DIR  -DBUILD_SHARED_LIBS=$buildsharedlibs -DCMAKE_BUILD_TYPE=$buildtype -DITK_USE_SYSTEM_PNG=ON -DITK_USE_SYSTEM_TIFF=ON -DITK_USE_SYSTEM_ZLIB=ON
          try make $makeflags
          try make install
          # Ensure that the insight installation suceeded.
          try cd $SCIRUN_THIRDPARTY_DIR/lib/InsightToolkit
          try mkdir -p $SCIRUN_THIRDPARTY_DIR/include/InsightToolkit/itkExtHdrs
          try cp $DIR/$INSIGHT_TOOLKIT/Utilities/itkExtHdrs/itkAnalyzeDbh.h $SCIRUN_THIRDPARTY_DIR/include/InsightToolkit/itkExtHdrs/itkAnalyzeDbh.h
        else
          # Cannot install programs with cmake under OSX as libraries
          # are improperly linked using cmake
          
          cd $DIR/$INSIGHT_TOOLKIT
          
          if [ "$darwin64" == "1" ]; then
            if [ "$universal" == "1" ]; then
              $cmakebin . -DCMAKE_C_FLAGS="-arch x86_64 -arch ppc64 " -DCMAKE_CXX_FLAGS="-arch x86_64 -arch ppc64 " -DCMAKE_EXE_LINKER_FLAGS="-arch x86_64 -arch ppc64 " -DBUILD_EXAMPLES=0 -DBUILD_TESTING=0 -DBUILD_SHARED_LIBS=$buildsharedlibs -DCMAKE_BUILD_TYPE=$buildtype
            else
              $cmakebin . -DCMAKE_C_FLAGS="-m64" -DCMAKE_CXX_FLAGS="-m64" -DCMAKE_EXE_LINKER_FLAGS="-m64" -DBUILD_EXAMPLES=0 -DBUILD_TESTING=0 -DBUILD_SHARED_LIBS=$buildsharedlibs -DCMAKE_BUILD_TYPE=$buildtype
            fi
          else
            if [ "$universal" == "1" ]; then
              $cmakebin . -DCMAKE_C_FLAGS="-arch i386 -arch ppc" -DCMAKE_CXX_FLAGS="-arch i386 -arch ppc" -DCMAKE_EXE_LINKER_FLAGS="-arch i386 -arch ppc" -DBUILD_EXAMPLES=0 -DBUILD_TESTING=0 -DBUILD_SHARED_LIBS=$buildsharedlibs -DCMAKE_BUILD_TYPE=$buildtype
            else
              $cmakebin . -DBUILD_EXAMPLES=0 -DBUILD_TESTING=0 -DBUILD_SHARED_LIBS=$buildsharedlibs -DCMAKE_BUILD_TYPE=$buildtype
            fi
          fi          
          
          try make $makeflags VERBOSE=1
          cmakeargs="${cmakeargs} -DITK_DIR=$DIR/$INSIGHT_TOOLKIT"          
        fi
    fi
}


build_wxWidgets() {
    if [ $buildwx == "1" ]; then
        # Set up which one to build based upon architecture.
        WXWIDGETS_TOOLKIT=$WXWIDGETS_TOOLKIT_GTK
        WXWIDGETS_TOOLKIT_ARG=$WXWIDGETS_TOOLKIT_ARG_GTK
        if [ "$osx" == "1" ]; then
            WXWIDGETS_TOOLKIT=$WXWIDGETS_TOOLKIT_MAC
            WXWIDGETS_TOOLKIT_ARG=$WXWIDGETS_TOOLKIT_ARG_MAC
        fi
        try cd $DIR
        if [ ! -e "$DIR/${WXWIDGETS_TOOLKIT}.tar.gz" ]; then 
            try echo "Downloading wxWidgets..."
            try $getcommand http://downloads.sourceforge.net/wxwindows/${WXWIDGETS_TOOLKIT}.tar.gz
        fi
        
        if [ ! -e "$DIR/$WXWIDGETS_TOOLKIT" ]; then
            try tar xvzf ${WXWIDGETS_TOOLKIT}.tar.gz
        fi
        
        try cd $DIR/$WXWIDGETS_TOOLKIT
        try ./configure $WXWIDGETS_TOOLKIT_ARG --with-opengl --disable-shared --prefix=$SCIRUN_THIRDPARTY_DIR
        try make $makeflags
        try make install

        # Ensure that the wxwidgets installation suceeded
        try cd $SCIRUN_THIRDPARTY_DIR/lib/wx
        cmakewxargs="-DWITH_WXWIDGETS=ON -DwxWidgets_wxrc_EXECUTABLE=${SCIRUN_THIRDPARTY_DIR}/bin/wxrc -DwxWidgets_CONFIG_EXECUTABLE=${SCIRUN_THIRDPARTY_DIR}/bin/wx-config"
    fi
}


find_svn() {
    # Try to find a version of svn
    svnbin=`which svn`
    echo svnbin=$svnbin
}

update_scirun_src() {
    echo "Updating SCIRUN source..."

    # If the source tree is from SVN, update it too
    #if [ -e "$DIR/src/.svn" ] && [ -e "$svnbin" ]; then 
    #    try svn update "$DIR/src"
    #fi
}
    
configure_scirun_bin() {
    try cd $DIR/bin
    
    if [ "$builddataflow" == "1" ]; then
      buildsharedlibs="1"
    fi

    # Put common SCIRun build options here:
    local COMMON_BUILD_OPTS="-DBUILD_SHARED_LIBS=$buildsharedlibs \
                             -DBUILD_SEG3D=$buildseg3d \
                             -DBUILD_DATAFLOW=$builddataflow \
                             -DCMAKE_BUILD_TYPE=$buildtype \
                             -DLOAD_PACKAGE=$scirunpackages \
                             -DWITH_X11=$withx11 \
                             -DBUILD_TESTING=$testnetworks \
                             -DBUILD_UTILS=$buildutils \
                             -DBUILD_BIOMESH3D=$buildmeshing \
                             -DWITH_TETGEN=$buildtetgen \
                             -DWITH_LAPACK=$buildlapack"

    if [ "$osx" == "1" ]; then
      if [ "$darwin64" == "1" ]; then
        if [ "$universal" == "1" ]; then
          $cmakebin ../src -DSCIRUN_ENABLE_64BIT=ON -DCMAKE_C_FLAGS="-arch x86_64 -arch ppc64 " -DCMAKE_CXX_FLAGS="-arch x86_64 -arch ppc64 " -DCMAKE_EXE_LINKER_FLAGS="-arch x86_64 -arch ppc64 " $COMMON_BUILD_OPTS $cmakewxargs $cmakeargs
        else
            $cmakebin ../src -DSCIRUN_ENABLE_64BIT=ON -DCMAKE_C_FLAGS="-m64 " -DCMAKE_CXX_FLAGS="-m64 " -DCMAKE_EXE_LINKER_FLAGS="-m64 " $COMMON_BUILD_OPTS $cmakewxargs $cmakeargs
        fi
      else
        if [ "$universal" == "1" ]; then
          $cmakebin ../src -DCMAKE_C_FLAGS="-arch i386 -arch ppc " -DCMAKE_CXX_FLAGS="-arch i386 -arch ppc " -DCMAKE_EXE_LINKER_FLAGS:="-arch i386 -arch ppc " $COMMON_BUILD_OPTS $cmakewxargs $cmakeargs
        else
          $cmakebin ../src $COMMON_BUILD_OPTS $cmakewxargs $cmakeargs
        fi
      fi
    else
      $cmakebin ../src $COMMON_BUILD_OPTS $cmakewxargs $cmakeargs
    fi
}

build_scirun_bin() {
    if [ "$usectest" == "1" ] && [ -e "$ctestbin" ]; then
        echo "Building SCIRun using ctest..."
        trybuild $ctestbin -VV -D Experimental -A $DIR/bin/CMakeCache.txt
    else 
	echo "Building SCIRun using make..."
        trybuild make $makeflags
    fi
}


######### build.sh script execution starts here

export DIR=`pwd`
export SCIRUN_THIRDPARTY_DIR=$DIR/bin

linux=0
osx=0

if test `uname` = "Darwin"; then
    getcommand="curl -OL"
    osx=1
elif test `uname` = "Linux"; then
    getcommand="wget"
    linux=1
else
    echo "Unsupported system.  Please run on OSX or Linux"
    exit 1
fi

buildtype="Release"
buildsharedlibs="1"
builddataflow="1"
buildseg3d="0"
makeflags=""
cmakeflags=""
builditk=0
buildwx=0
buildmeshing="1"
buildtetgen="1"
buildlapack="0"
cmakewxargs=""
getcmake=0
withx11="1"
scirunpackages="SCIRun,BioPSE,Teem,MatlabInterface"
cmakeargs=""
usectest="0"
testnetworks="0"
buildutils="1"
darwin64="0"
universal="0"
documentation="0"

echo "Parsing arguments..."
while [ "$1" != "" ]; do
    case "$1" in
        --universal)
            universal="1";;
        --documentation)
            documentation="1";;
        --debug)
            buildtype="Debug";;
        --release)
            buildtype="Release";;
        --with-itk) 
            builditk="1";;
        --with-wxwidgets) 
            buildwx="1";;
        --without-tetgen)
            buildtetgen="0";;
        --with-lapack)
            buildlapack="1";;
        --without-meshing-pipeline)
            buildmeshing="0";;
        --get-cmake) 
            getcmake="1";;
        --no-ctest) 
            usectest="0";;
        --ctest) 
            usectest="1";;
        --test-networks)
            testnetworks="1";;
        --static) 
            buildsharedlibs="0";;
        --seg3d-only) 
            buildseg3d="1"
            builddataflow="0"
            buildsharedlibs="0"
            buildwx="1"
            builditk="1"
            buildutils="1"
            if [ "$osx" == "1" ]; then
                withx11="0"
            fi;;
       --64bit)
            if [ "$osx" == "1" ]; then
              darwin64="1"
                      else
              echo "Only OSX supporsts the selection between 32 and 64bit architectures."
              echo "On other platforms SCIRun uses the architecture of the operating system."
            fi;;
       --packages=*)
            scirunpackages=`echo $1 | awk -F '\=' '{print $2}'`;;
        --cmake-args=*)
            cmakeargs=`echo $1 | cut -c 14-`;;
        -j*) 
            makeflags="${makeflags} $1";;
        -D*) 
            cmakeflags="${cmakeflags} $1";;
        -?|--?|-help|--help) 
            printhelp;;
        *) 
            echo \`$1\' parameter ignored;;
    esac
    shift 1
done

cmakeargs="${cmakeargs} ${cmakeflags}"

if [ "$documentation" != "0" ] ; then
  cmakeargs="${cmakeargs} -DBUILD_DOCUMENTATION=ON"
fi

echo "CMake args: $cmakeargs"
echo "Build Type: $buildtype"
echo "Build SCIRun Dataflow: $builddataflow"
if [ "$builddataflow" != "0" ]; then
    echo "SCIRun Packages: $scirunpackages"
fi
echo "Build Insight ToolKit: $builditk"
echo "Build wxWidgets: $buildwx"
if [ "$buildseg3d" != "0" ] ; then
    echo "Build Seg3D"
fi
echo "Get CMake: $getcmake"
echo "Make Flags: $makeflags"
echo "Build Shared Libs: $buildsharedlibs"
echo "With X11: $withx11"
echo "With LAPACK: $buildlapack"

ensure make --version # ensure make is on the system

find_cmake

find_svn

build_ITK

build_wxWidgets

update_scirun_src

configure_scirun_bin

build_scirun_bin
