#!/usr/bin/env python
#  
#  For more information, please see: http://software.sci.utah.edu
#  
#  The MIT License
#  
#  Copyright (c) 2009 Scientific Computing and Imaging Institute,
#  University of Utah.
#  
#  
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#  
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#  
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#  
#    Author : Martin Cole

import string
from sys import argv
import sys
import subprocess
import re
import os
import time
import glob
import shutil
from thread import start_new_thread
from math import fabs

print_only = False
use_shell = True
if sys.platform == "win32" :
    use_shell = False

def dump_log(p) :
    if print_only :
        return

    # Warning: The data read is buffered in memory, so do not use this method if the data size is large or unlimited.
    # See Python docs, section 18.1.2.
    (stdoutdata, stderrdata) = p.communicate()
    f = open("%d.log" % p.pid, "w")
    f.writelines(stdoutdata)
    f.close()

def wait_on_procs(procs) :    

    # the output from processes need to be dumped or the
    # buffer fills up and it sleeps.
    for p in procs :
        print "starting thread to dump"
        print p
        start_new_thread(dump_log, (p, ))
        

    #once all tight procs are done, make the vols    
    done = print_only
    count = 0;
    while not done :
        time.sleep(1)
        ndone = 0
        #all tight procs need to be finished before moving on.
        np = 0
        for p in procs :
            if p.poll() != None : #None means process not finished.
                ndone = ndone + 1

        if ndone == len(procs) :
            done = True
        else :
            if count % 30 == 0 :
                print "%d processes are complete of %d" % (ndone, len(procs))
        count = count + 1


def assert_exists(file):
    if not(os.path.exists(file)):
        print "No such file %s. Perhaps you need to run a previous stage."%file
        sys.exit(-1)

def do_system(cmmd) :
    if print_only :
        print cmmd
    else :
        ret = -1
        try:
            ret = subprocess.call(cmmd, shell=use_shell)
            if (ret < 0):
                sys.exit(ret)
        except OSError, e:
            print "Execution failed with error ", e
            sys.exit(ret)

def gen_isect_file(fname) :
    global g_num_isects
    global g_nquads
    global g_ntrips
    global g_ndoubs
    global g_quad_tri_juncs
    global g_quad_tri_seeds

    print "test: in gen_isect_file ", fname

    g_num_isects = 0
    g_quad_tri_juncs = []
    g_quad_tri_seeds = []

    seed_path = "seeds"
    if not os.path.exists(seed_path):
        print "Path %s does not exist.  Perhaps you need to run a previous stage." %seed_path
        sys.exit(-1)

    input_seed_path = "input_seeds"
    if not os.path.exists(input_seed_path):
        os.mkdir(input_seed_path)

    isect_file_text = ""

    nmats = len(mats)

    # first write out the quad junctions to be optimized
    g_nquads = 0
    quad_text = ""

    for i in range(0,nmats-3) :
        for j in range(i+1,nmats-2) :
            for k in range(j+1,nmats-1) :
                for l in range(k+1,nmats) :
                    seed_file = "%s/q%s%s%s%s_seed.ptcl" % (seed_path, i, j, k, l)
                    if os.path.exists(seed_file):
                        input_seed_file = "%s/q%s%s%s%s_seed.ptcl" % (input_seed_path, i, j, k, l)
                        shutil.copyfile(seed_file, input_seed_file)
                        g_quad_tri_seeds.append(input_seed_file)
                        g_quad_tri_juncs.append("junctions/particle_params.txt_%s.ptcl" % (g_nquads))
                        quad_text += "%s %s %s %s y %s y\n" % (i, j, k, l, input_seed_file)
                        g_nquads+=1

    isect_file_text += "%s\n" % g_nquads + quad_text

    # second write out the triple junctions to be optimized
    g_ntrips = 0
    trip_text = ""

    for i in range(0,nmats-2) :
        for j in range(i+1,nmats-1) :
            for k in range(j+1,nmats) :
                seed_file = "%s/t%s%s%s_seed.ptcl" % (seed_path,i,j,k)
                if os.path.exists(seed_file):
                    input_seed_file = "%s/t%s%s%s_seed.ptcl" % (input_seed_path, i, j, k)
                    shutil.copyfile(seed_file, input_seed_file)
                    g_quad_tri_seeds.append(input_seed_file)
                    g_quad_tri_juncs.append("junctions/particle_params.txt_%s.ptcl" % (g_nquads+g_ntrips))
                    trip_text += "%s %s %s y %s y\n" % (i, j, k, input_seed_file)
                    g_ntrips+=1

    isect_file_text += "%s\n" % g_ntrips + trip_text

    # finally write out the double junctions
    g_ndoubs = 0
    doub_text = ""

    for i in range(0,nmats-1) :
        for j in range(i+1,nmats) :
            seed_file = "%s/d%s%s_seed.ptcl" % (seed_path, i, j)
            if os.path.exists(seed_file):
                input_seed_file = "%s/d%s%s_seed.ptcl" % (input_seed_path, i, j)
                shutil.copyfile(seed_file, input_seed_file)
                g_ndoubs+=1
                doub_text += "%s %s y %s y\n" % (i, j, input_seed_file)

    isect_file_text += "%s\n" % g_ndoubs + doub_text

    f = open(fname, 'w')
    f.write(isect_file_text)
    f.close()

    g_num_isects = g_nquads+g_ntrips+g_ndoubs 

if __name__ == "__main__" :

# Check that we got the right arguments......
    if len(argv) < 2 or not(os.path.exists(argv[1])):
        print 'Usage: %s [model_config] [binary_path]' % argv[0];
        exit(1)

    MAX_SIZING_FIELD = ""
    SIZING_SCALE_VAR = ""

    model_config = argv[1]
    execfile(model_config)

    model_path, dummy = os.path.split(model_config)
    
    if not(os.path.isabs(model_input_file)) :
      model_input_file = os.path.normpath(os.path.join(model_path, model_input_file))

    if not(os.path.isabs(model_output_path)) :
      model_output_path = os.path.normpath(os.path.join(model_path, model_output_path))
   
    
    if len(argv) > 2:
        binary_path = argv[2]
        if binary_path[len(binary_path) - 1] != os.sep:
            binary_path += os.sep
    else:
        binary_path = ""  

    #sys.path.append('%s/lib' % scirun_bin)

# Done checking arguments

    cur_dir = os.getcwd();
    os.chdir(model_output_path);

    start_time = time.time()

    procs = []
    dnm = "junctions"

    cwd = os.getcwd()
    if not os.path.exists(dnm) :
        os.mkdir(dnm)

    gen_isect_file( "%s/m1.txt" % dnm )

    idx = 0
    lns = ["%d\n" % len(mats)]
    for i in mats :
        n = mat_names[idx]
        lns.append("%s/%s\n" % (cwd, n))
        idx = idx + 1

    f = open('%s/particle_params.txt' % dnm, 'w')
    f.writelines(lns)
    f.close()

    if MAX_SIZING_FIELD != "" :
        max_sizing_field = MAX_SIZING_FIELD
      

    psys_txt = [
            "ENERGY                  radial\n",
            "NUMBER_OF_SURFACES      %d\n" % len(mats),
            "NUMBER_OF_INTERSECTIONS %d\n" % g_num_isects,
            "INTERSECTION_FILE       %s/%s/m1.txt\n" % (cwd, dnm),
            "BASE_FILE_NAME          %s/%s/particle_params.txt\n" % (cwd,dnm),
            "INIT_NUM_POINTS         5\n", #unused
            "MAX_SF                  %f\n" % max_sizing_field,
            "SIZING_SCALE            %f\n" % 0.4
            ]

    f = open('%s/psystem_input.txt' % dnm, 'w')
    f.writelines(psys_txt)
    f.close()

    # now run rendersf3d on the resulting junctions directory
    optimize_ps_cmmd = r'"%soptimize-particle-system" junctions%spsystem_input.txt %s 0 %s' % (binary_path, os.sep, num_particle_iters, g_nquads+g_ntrips-1)
#    optimize_ps_cmmd = "%s/bin/optimize-particle-system junctions/psystem_input.txt %s 0 %s" % (scirun_thirdparty,num_particle_iters,g_nquads+g_ntrips+g_ndoubs)
    print "**** running command: %s\n" % optimize_ps_cmmd
    do_system(optimize_ps_cmmd)

    # Now copy the junctions results for the quads and triples back to the
    # input_seed_path files (but need to map names!)
    # And then fork a bunch of processes -- once for each double
    # and call optimize_ps_cmd just for that interface #

    print "**** copying quad and tri junctions back into seeds folder\n"
    for i in range(0, g_nquads+g_ntrips) :
        shutil.copyfile(g_quad_tri_juncs[i], g_quad_tri_seeds[i])
    
    procs = []
    for i in range(g_nquads+g_ntrips, g_nquads+g_ntrips+g_ndoubs) :
        optimize_ps_cmmd = r'"%soptimize-particle-system" junctions%spsystem_input.txt %s %s %s >ops-output-%s.txt' % (binary_path, os.sep, num_particle_iters, i, i, i)
        print "**** parallel process %s of %s: %s\n"%(i-g_nquads-g_ntrips+1,g_ndoubs,optimize_ps_cmmd)
        procs.append(subprocess.Popen(optimize_ps_cmmd,  shell=True, bufsize=0, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=model_output_path))

    wait_on_procs(procs)

    # change directory to junctions and write out files for visualization
    os.chdir("junctions")
    print "**** Change directory"

    for f in (glob.glob('./particle_params.txt_*.ptcl')):
        (params_base, params_ext) = os.path.splitext(f)
        ptcl_data_cmmd = r'"%sPtclToField" %s %s.pcv.fld' % (binary_path, f, params_base)
        do_system(ptcl_data_cmmd)
        print "**** doing command %s" % ptcl_data_cmmd
        ptcl_nodata_cmmd = r'"%sPtclToField" -nodata %s %s.pc.fld' % (binary_path, f, params_base)
        do_system(ptcl_nodata_cmmd)
        print "**** doing command %s" % ptcl_nodata_cmmd

    if os.path.exists("particle-all.pc.fld"):
        os.remove("particle-all.pc.fld")
    join_fields_nodata_cmmd = r'"%sJoinFields" -indexdata particle-all.pc.fld %s' % (binary_path, string.join(glob.glob('./particle_params.txt_*.pc.fld'), ' '))
    do_system(join_fields_nodata_cmmd)

    if os.path.exists("particle-all.pcv.fld"):
        os.remove("particle-all.pcv.fld")
    join_fields_data_cmmd = r'"%sJoinFields" particle-all.pcv.fld %s' % (binary_path, string.join(glob.glob('./particle_params.txt_*.pcv.fld'), ' '))
    do_system(join_fields_data_cmmd)

    os.chdir("..")

    stop_time = time.time()
    f = open("distribute-particles-runtime.txt","w")
    f.write("%lf" % (stop_time-start_time))
    f.close()

    os.chdir(cur_dir)
