/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2009 Scientific Computing and Imaging Institute,
   University of Utah.

   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/



/*
 *  GeomArrows.cc: Arrows object
 *
 *  Written by:
 *   Steven G. Parker
 *   Department of Computer Science
 *   University of Utah
 *   June 1995
 *
 */

#include <Core/Util/Debug.h>

#include <Core/Geom/GeomArrows.h>
#include <Core/Geom/DrawInfoOpenGL.h>

#include <Core/Math/MiscMath.h>

namespace SCIRun {

GeomArrows::GeomArrows(double headwidth, double headlength, int cyl, double r,
		       int normhead)
  : headwidth(headwidth), headlength(headlength), rad(r), drawcylinders(cyl),
    normalize_headsize(normhead)
{
  DEBUG_CONSTRUCTOR("GeomArrows")

  shaft_matls.add(new Material(Color(0,0,0), Color(.6, .6, .6), Color(.6, .6, .6), 10));
  head_matls.add(new Material(Color(0,0,0), Color(1,0,0), Color(.6, .6, .6), 10));
  back_matls.add(new Material(Color(0,0,0), Color(.6, .6, .6), Color(.6, .6, .6), 10));
}

GeomArrows::GeomArrows(const GeomArrows& copy)
: GeomObj(copy)
{
  DEBUG_CONSTRUCTOR("GeomArrows")
}

GeomArrows::~GeomArrows() 
{
  DEBUG_DESTRUCTOR("GeomArrows")
}

void 
GeomArrows::set_material(const MaterialHandle& shaft_matl,
                         const MaterialHandle& back_matl,
                         const MaterialHandle& head_matl)
{
  shaft_matls.resize(1);
  back_matls.resize(1);
  head_matls.resize(1);
  shaft_matls[0]=shaft_matl;
  back_matls[0]=back_matl;
  head_matls[0]=head_matl;
}

void 
GeomArrows::add(const Point& pos, const Vector& dir,
                const MaterialHandle& shaft, const MaterialHandle& back,
                const MaterialHandle& head)
{
  add(pos, dir);
  shaft_matls.add(shaft);
  back_matls.add(back);
  head_matls.add(head);
}

void
GeomArrows::add(const Point& pos, const Vector& dir)
{
  Vector vv1, vv2;
  if(!dir.check_find_orthogonal(vv1, vv2))
    return;

  positions.add(pos);
  directions.add(dir);
  if (!normalize_headsize) 
  {
    // use the length to scale the head
    double len = dir.length();
    v1.add(vv1*headwidth*len);
    v2.add(vv2*headwidth*len);
  } 
  else 
  {
    // don't scale the head by the length
    vv1.normalize();
    vv2.normalize();
    v1.add(vv1*headwidth);
    v2.add(vv2*headwidth);
  }
}

void 
GeomArrows::get_bounds(BBox& bb)
{
  int n=positions.size();
  for(int i=0;i<n;i++)
  {
    bb.extend(positions[i]);
    bb.extend(positions[i]+directions[i]);
  }
}

GeomObj* 
GeomArrows::clone()
{
  return new GeomArrows(*this);
}

void
GeomArrows::draw(DrawInfoOpenGL* di, Material* matl, double)
{
  int n = positions.size();
  di->polycount_ += 6 * n;

  // Draw shafts - they are the same for all draw types.
  double shaft_scale = headlength;

  // if we're not drawing cylinders, draw lines
  if ( drawcylinders == 0 )
  {
    if (di->get_drawtype() == DrawInfoOpenGL::WireFrame)
      shaft_scale=1.0;
    if (shaft_matls.size() == 1)
    {
      if (!pre_draw(di, shaft_matls[0].get_rep(), 0)) return;
      glBegin(GL_LINES);
      for (int i=0;i<n;i++)
      {
        Point from(positions[i]);
        Point to(from+directions[i]*shaft_scale);
        glVertex3d(from.x(), from.y(), from.z());
        glVertex3d(to.x(), to.y(), to.z());
      }
      glEnd();
    }
    else
    {
      if (!pre_draw(di, matl, 0)) return;
      glBegin(GL_LINES);
      for (int i=0;i<n;i++)
      {
        di->set_material(shaft_matls[i+1].get_rep());
        Point from(positions[i]);
        Point to(from+directions[i]*shaft_scale);
        glVertex3d(from.x(), from.y(), from.z());
        glVertex3d(to.x(), to.y(), to.z());
      }
      glEnd();
    }

  }
  else
  {
    // number of subdivisions
    int nu = 4;
    int nv = 1;
        
    // drawing cylinders
    if ( shaft_matls.size() == 1)
    {
      if (!pre_draw(di, shaft_matls[0].get_rep(), 1)) return;
      for ( int i =0; i < n; i++ )
      {
        Point from(positions[i]);
        Point to(from+directions[i]*shaft_scale);
        // create cylinder along axis with endpoints from and to
        Vector axis = to - from;
        Vector z(0,0,1);
        Vector zrotaxis;
        double zrotangle;
        if ( Abs(axis.y())+Abs(axis.x()) < 1.e-5)
        {
          // Only in x-z plane.
          zrotaxis=Vector(0,-1,0);
        }
        else
        {
          zrotaxis=Cross(axis, z);
          zrotaxis.normalize();
        }
        double cangle=Dot(z, axis)/axis.length();
        zrotangle=-Acos(cangle);

        // draw cylinder
        glPushMatrix();
        glTranslated( from.x(), from.y(), from.z() );
        glRotated( RtoD(zrotangle), zrotaxis.x(), zrotaxis.y(), zrotaxis.z());
        di->polycount_ += 2*(nu-1)*(nv-1);
        gluCylinder(di->qobj_, rad, rad, axis.length(), nu, nv);
        glPopMatrix();
      }
    }
    else
    {
      if (!pre_draw(di, matl, 1)) return;
      for ( int i =0; i < n; i++ )
      {
        di->set_material(shaft_matls[i+1].get_rep());
        Point from(positions[i]);
        Point to(from+directions[i]*shaft_scale);
        // create cylinder along axis with endpoints from and to
        Vector axis = to - from;
        Vector z(0,0,1);
        Vector zrotaxis;
        double zrotangle;
        if ( Abs(axis.y())+Abs(axis.x()) < 1.e-5)
        {
          // Only in x-z plane.
          zrotaxis=Vector(0,-1,0);
        }
        else
        {
          zrotaxis=Cross(axis, z);
          zrotaxis.normalize();
        }
        double cangle=Dot(z, axis)/axis.length();
        zrotangle=-Acos(cangle);

        // draw cylinder
        glPushMatrix();
        glTranslated( from.x(), from.y(), from.z() );
        glRotated( RtoD(zrotangle), zrotaxis.x(), zrotaxis.y(), zrotaxis.z());
        di->polycount_ += 2*(nu-1)*(nv-1);
        gluCylinder(di->qobj_, rad, rad, axis.length(), nu, nv);
        glPopMatrix();
      }
    }
  }

  if (headwidth == 0 || headlength == 0) { post_draw(di); return; }

  // Draw back and head
  switch(di->get_drawtype())
  {
  case DrawInfoOpenGL::WireFrame:
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    break;
  case DrawInfoOpenGL::Flat:
  case DrawInfoOpenGL::Gouraud:
    break;
  }

  int do_normals=1;
  if (di->get_drawtype() == DrawInfoOpenGL::Flat ||
      di->get_drawtype() == DrawInfoOpenGL::WireFrame)
    do_normals=0;

  if (drawcylinders == 0)
  {
    // Draw the back of the arrow
    if (back_matls.size() == 1)
    {
      if (!pre_draw(di, back_matls[0].get_rep(), 1)) return;
      glBegin(GL_QUADS);
      if (do_normals)
      {
        for (int i=0;i<n;i++)
        {
          glNormal3d(directions[i].x(), directions[i].y(), directions[i].z());
          Point from(positions[i]+directions[i]*headlength);
          Point p1(from+v1[i]);
          glVertex3d(p1.x(), p1.y(), p1.z());
          Point p2(from+v2[i]);
          glVertex3d(p2.x(), p2.y(), p2.z());
          Point p3(from-v1[i]);
          glVertex3d(p3.x(), p3.y(), p3.z());
          Point p4(from-v2[i]);
          glVertex3d(p4.x(), p4.y(), p4.z());
        }
      }
      else
      {
        for (int i=0;i<n;i++)
        {
          Point from(positions[i]+directions[i]*headlength);
          Point p1(from+v1[i]);
          glVertex3d(p1.x(), p1.y(), p1.z());
          Point p2(from+v2[i]);
          glVertex3d(p2.x(), p2.y(), p2.z());
          Point p3(from-v1[i]);
          glVertex3d(p3.x(), p3.y(), p3.z());
          Point p4(from-v2[i]);
          glVertex3d(p4.x(), p4.y(), p4.z());
        }
      }
      glEnd();
    }
    else
    {
      if (!pre_draw(di, matl, 1)) return;
      glBegin(GL_QUADS);
      if (do_normals)
      {
        for (int i=0;i<n;i++)
        {
          di->set_material(back_matls[i+1].get_rep());
          glNormal3d(directions[i].x(), directions[i].y(), directions[i].z());
          Point from(positions[i]+directions[i]*headlength);
          Point p1(from+v1[i]);
          glVertex3d(p1.x(), p1.y(), p1.z());
          Point p2(from+v2[i]);
          glVertex3d(p2.x(), p2.y(), p2.z());
          Point p3(from-v1[i]);
          glVertex3d(p3.x(), p3.y(), p3.z());
          Point p4(from-v2[i]);
          glVertex3d(p4.x(), p4.y(), p4.z());
        }
      }
      else
      {
        for (int i=0;i<n;i++)
        {
          di->set_material(back_matls[i+1].get_rep());
          Point from(positions[i]+directions[i]*headlength);
          Point p1(from+v1[i]);
          glVertex3d(p1.x(), p1.y(), p1.z());
          Point p2(from+v2[i]);
          glVertex3d(p2.x(), p2.y(), p2.z());
          Point p3(from-v1[i]);
          glVertex3d(p3.x(), p3.y(), p3.z());
          Point p4(from-v2[i]);
          glVertex3d(p4.x(), p4.y(), p4.z());
        }
      }
      glEnd();
    }

    // Draw the head of the arrow
    if (head_matls.size() == 1)
    {
      if (!pre_draw(di, head_matls[0].get_rep(), 1)) return;
      if (do_normals)
      {
        double w=headwidth;
        double h=1.0-headlength;
        double w2h2=w*w/h;
        for (int i=0;i<n;i++)
        {
          glBegin(GL_TRIANGLES);
          Vector dn(directions[i]*w2h2);
          Vector n(dn+v1[i]+v2[i]);
          glNormal3d(n.x(), n.y(), n.z());
          Point top(positions[i]+directions[i]);
          Point from;
          if (normalize_headsize == 0 )
          {
            from = top - directions[i] * h;
          }
          else
          {
            Vector dir( directions[i] );
            dir.normalize();
            from = top - dir * h;
          }
          Point p1(from+v1[i]);
          Point p2(from+v2[i]);
          glVertex3d(top.x(), top.y(), top.z());
          glVertex3d(p1.x(), p1.y(), p1.z());
          glVertex3d(p2.x(), p2.y(), p2.z()); // 1st tri
          n=dn-v1[i]+v2[i];
          glNormal3d(n.x(), n.y(), n.z());
          Point p3(from-v1[i]);
          glVertex3d(top.x(), top.y(), top.z());
          glVertex3d(p2.x(), p2.y(), p2.z());
          glVertex3d(p3.x(), p3.y(), p3.z()); // 2nd tri
          n=dn-v1[i]-v2[i];
          glNormal3d(n.x(), n.y(), n.z());
          Point p4(from-v2[i]);
          glVertex3d(top.x(), top.y(), top.z());
          glVertex3d(p3.x(), p3.y(), p3.z());
          glVertex3d(p4.x(), p4.y(), p4.z()); // 3rd tri
          n=dn+v1[i]-v2[i];
          glNormal3d(n.x(), n.y(), n.z());
          glVertex3d(top.x(), top.y(), top.z());
          glVertex3d(p4.x(), p4.y(), p4.z());
          glVertex3d(p1.x(), p1.y(), p1.z()); // 4th tri
          glEnd();
        }
      }
      else
      {
        for (int i=0;i<n;i++)
        {
          glBegin(GL_TRIANGLE_FAN);
          Point from(positions[i]+directions[i]);
          glVertex3d(from.x(), from.y(), from.z());
          if (normalize_headsize == 0 ) from-=directions[i]*(1.0-headlength);
          else
          {
            Vector dir( directions[i] );
            dir.normalize();
            from -= dir*(1.0-headlength);
          }
          Point p1(from+v1[i]);
          glVertex3d(p1.x(), p1.y(), p1.z());
          Point p2(from+v2[i]);
          glVertex3d(p2.x(), p2.y(), p2.z());
          Point p3(from-v1[i]);
          glVertex3d(p3.x(), p3.y(), p3.z());
          Point p4(from-v2[i]);
          glVertex3d(p4.x(), p4.y(), p4.z());
          glVertex3d(p1.x(), p1.y(), p1.z());
          glEnd();
        }
      }
    }
    else
    {
      if (!pre_draw(di, matl, 1)) return;
      if (do_normals)
      {
        double w=headwidth;
        double h=1.0-headlength;
        double w2h2=w*w/h;
        for (int i=0;i<n;i++)
        {
          glBegin(GL_TRIANGLES);
          di->set_material(head_matls[i+1].get_rep());
          Vector dn(directions[i]*w2h2);
          Vector n(dn+v1[i]+v2[i]);
          glNormal3d(n.x(), n.y(), n.z());
          Point top(positions[i]+directions[i]);
          Point from;
          if (normalize_headsize == 0 )
          {
            from = top - directions[i] * h;
          }
          else
          {
            Vector dir( directions[i] );
            dir.normalize();
            from = top - dir * h;
          }
          Point p1(from+v1[i]);
          Point p2(from+v2[i]);
          glVertex3d(top.x(), top.y(), top.z());
          glVertex3d(p1.x(), p1.y(), p1.z());
          glVertex3d(p2.x(), p2.y(), p2.z()); // 1st tri
          n=dn-v1[i]+v2[i];
          glNormal3d(n.x(), n.y(), n.z());
          Point p3(from-v1[i]);
          glVertex3d(top.x(), top.y(), top.z());
          glVertex3d(p2.x(), p2.y(), p2.z());
          glVertex3d(p3.x(), p3.y(), p3.z()); // 2nd tri
          n=dn-v1[i]-v2[i];
          glNormal3d(n.x(), n.y(), n.z());
          Point p4(from-v2[i]);
          glVertex3d(top.x(), top.y(), top.z());
          glVertex3d(p3.x(), p3.y(), p3.z());
          glVertex3d(p4.x(), p4.y(), p4.z()); // 3rd tri
          n=dn+v1[i]-v2[i];
          glNormal3d(n.x(), n.y(), n.z());
          glVertex3d(top.x(), top.y(), top.z());
          glVertex3d(p4.x(), p4.y(), p4.z());
          glVertex3d(p1.x(), p1.y(), p1.z()); // 4th tri
          glEnd();
        }
      }
      else
      {
        for (int i=0;i<n;i++)
        {
          glBegin(GL_TRIANGLE_FAN);
          di->set_material(head_matls[i+1].get_rep());
          Point from(positions[i]+directions[i]);
          glVertex3d(from.x(), from.y(), from.z());
          from-=directions[i]*(1.0-headlength);
          if (normalize_headsize == 0 )
          {
            from -= directions[i] * (1.0 - headlength);
          }
          else
          {
            Vector dir( directions[i] );
            dir.normalize();
            from -= dir*(1.0-headlength);
          }
          Point p1(from+v1[i]);
          glVertex3d(p1.x(), p1.y(), p1.z());
          Point p2(from+v2[i]);
          glVertex3d(p2.x(), p2.y(), p2.z());
          Point p3(from-v1[i]);
          glVertex3d(p3.x(), p3.y(), p3.z());
          Point p4(from-v2[i]);
          glVertex3d(p4.x(), p4.y(), p4.z());
          glVertex3d(p1.x(), p1.y(), p1.z());
          glEnd();
        }
      }
    }

    switch(di->get_drawtype())
    {
    case DrawInfoOpenGL::WireFrame:
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
      break;
    case DrawInfoOpenGL::Flat:
    case DrawInfoOpenGL::Gouraud:
      break;
    }
    post_draw(di);

  }
  else
  {
    // Draw the head as a capped cone
    int nu = 4;
    int nv = 1;
    int nvdisk = 1;

    if (head_matls.size() == 1)
    {
      if (!pre_draw(di, head_matls[0].get_rep(), 1)) return;
      for (int i=0;i<n;i++)
      {
        glPushMatrix();
        Point top(positions[i]+directions[i]);
        Point bottom = top - directions[i]*(1-headlength);
        glTranslated(bottom.x(), bottom.y(), bottom.z());
        Vector axis = top - bottom;
        Vector z(0,0,1);
        Vector zrotaxis;
        double zrotangle;
        if ( Abs(axis.y())+Abs(axis.x()) < 1.e-5)
        {
          // Only in x-z plane.
          zrotaxis=Vector(0,-1,0);
        }
        else
        {
          zrotaxis=Cross(axis, z);
          zrotaxis.normalize();
        }
        const double cangle = Dot(z, axis)/axis.length();
        zrotangle = -Acos(cangle);
        glRotated(RtoD(zrotangle), zrotaxis.x(), zrotaxis.y(), zrotaxis.z());
        double bot_rad = 0.5 * headwidth;
        double top_rad = 0.01 * headwidth;
        di->polycount_ += 2 * (nu - 1) * (nv - 1);
        gluCylinder(di->qobj_, bot_rad, top_rad, headlength, nu, nv);
        if (bot_rad > 1.e-6)
        {
          // Bottom endcap
          di->polycount_ += 2*(nu-1)*(nvdisk-1);
          gluDisk(di->qobj_, 0, bot_rad, nu, nvdisk);
        }
        if (top_rad > 1.e-6)
        {
          // Top endcap
          glTranslated(0, 0, headlength);
          di->polycount_ += 2 * (nu - 1) * (nvdisk - 1);
          gluDisk(di->qobj_, 0, top_rad, nu, nvdisk);
        }
        glPopMatrix();
      }
    }
    else
    {
      if (!pre_draw(di, matl, 1)) return;
      for (int i=0;i<n;i++)
      {
        di->set_material(head_matls[i+1].get_rep());
        glPushMatrix();
        Point top(positions[i]+directions[i]);
        Point bottom = top - directions[i]*(1-headlength);
        glTranslated(bottom.x(), bottom.y(), bottom.z());
        Vector axis = top - bottom;
        Vector z(0,0,1);
        Vector zrotaxis;
        double zrotangle;
        if ( Abs(axis.y())+Abs(axis.x()) < 1.e-5)
        {
          // Only in x-z plane.
          zrotaxis=Vector(0,-1,0);
        }
        else
        {
          zrotaxis=Cross(axis, z);
          zrotaxis.normalize();
        }
        double cangle=Dot(z, axis)/axis.length();
        zrotangle=-Acos(cangle);
        glRotated(RtoD(zrotangle), zrotaxis.x(), zrotaxis.y(), zrotaxis.z());
        double bot_rad = 0.5*headwidth;
        double top_rad = 0.01*headwidth;
        di->polycount_+=2*(nu-1)*(nv-1);
        gluCylinder(di->qobj_, bot_rad, top_rad, headlength, nu, nv);
        if (bot_rad > 1.e-6)
        {
          // Bottom endcap
          di->polycount_+=2*(nu-1)*(nvdisk-1);
          gluDisk(di->qobj_, 0, bot_rad, nu, nvdisk);
        }
        if (top_rad > 1.e-6)
        {
          // Top endcap
          glTranslated(0, 0, headlength);
          di->polycount_+=2*(nu-1)*(nvdisk-1);
          gluDisk(di->qobj_, 0, top_rad, nu, nvdisk);
        }
        glPopMatrix();
      }
    }
    post_draw(di);
  }
}

} // End namespace SCIRun

// $Log

