/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2009 Scientific Computing and Imaging Institute,
   University of Utah.

   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/



/*
 *  GeomTexSlices.cc
 *
 *  Written by:
 *   Steven G. Parker
 *   Department of Computer Science
 *   University of Utah
 *   June 1995
 *
 */

#include <Core/Util/Debug.h>
#include <Core/Geom/DrawInfoOpenGL.h>
#include <Core/Geom/GeomTexSlices.h>

#ifdef _WIN32
#include <string.h>
#else
#include <strings.h>
#endif

namespace SCIRun {

GeomTexSlices::GeomTexSlices(int nx, int ny, int nz, const Point& min,
			     const Point &max)
  : accum(0.1),
    bright(0.6),
    min(min), max(max),
    nx(nx), ny(ny), nz(nz),
    have_drawn(0) 
{
  DEBUG_CONSTRUCTOR("GeomTexSlices")
  Xmajor.resize(nx, ny, nz);
  Ymajor.resize(ny, nx, nz);
  Zmajor.resize(nz, nx, ny);
}

GeomTexSlices::GeomTexSlices(const GeomTexSlices& copy)
  : GeomObj(copy)
{
  DEBUG_CONSTRUCTOR("GeomTexSlices")
}


GeomTexSlices::~GeomTexSlices()
{
  DEBUG_DESTRUCTOR("GeomTexSlices")
}

void GeomTexSlices::get_bounds(BBox& bb)
{
  bb.extend(min);
  bb.extend(max);
}

GeomObj* GeomTexSlices::clone()
{
  return new GeomTexSlices(*this);
}

void
GeomTexSlices::draw(DrawInfoOpenGL* di, Material* matl, double)
{
  if (!pre_draw(di, matl, 0)) return;
  if (!have_drawn)
  {
    have_drawn=1;
  }
  GLdouble model_mat[16]; // this is the modelview matrix

  glGetDoublev(GL_MODELVIEW_MATRIX,model_mat);

  // this is what you rip the view vector from
  // just use the "Z" axis, normalized

  Vector view = Vector(model_mat[0*4+2],model_mat[1*4+2],model_mat[2*4+2]);
  int sort_start=0;
  int sort_end=0;
  int sort_dir=1; // positive direction

  char which;

  if (Abs(view.x()) > Abs(view.y()))
  {
    if (Abs(view.x()) > Abs(view.z()))
    { // use x dir
      which = 0;
      if (view.x() < 0)
      {
        sort_dir=-1; sort_start=nx-1; sort_end=-1;
      }
      else
      {
        sort_start=0; sort_end=nx;
      }
    }
    else
    { // use z dir
      which = 2;
      if (view.z() < 0)
      {
        sort_dir=-1;sort_start =nz-1; sort_end=-1;
      }
      else
      {
        sort_start =0; sort_end=nz;
      }
    }
  }
  else if (Abs(view.y()) > Abs(view.z()))
  { // y greates
    which = 1;
    if (view.y() < 0)
    {
      sort_dir=-1;sort_start =ny-1; sort_end=-1;
    }
    else
    {
      sort_start =0; sort_end=ny;
    }
  }
  else
  { // z is the one
    which = 2;
    if (view.z() < 0)
    {
      sort_dir=-1;sort_start = nz-1; sort_end=-1;
    }
    else
    {
      sort_start = 0; sort_end = nz;
    }
  }

  Point pts[4];
  Vector v(0, 0, 0);
  switch (which)
  {
  case 0:       // x
    pts[0] = min;
    pts[1] = Point (min.x(), max.y(), min.z());
    pts[2] = Point (min.x(), max.y(), max.z());
    pts[3] = Point (min.x(), min.y(), max.z());
    v = Vector (max.x()-min.x(),0,0);
    break;
  case 1:       // y
    pts[0] = min;
    pts[1] = Point (max.x(), min.y(), min.z());
    pts[2] = Point (max.x(), min.y(), max.z());
    pts[3] = Point (min.x(), min.y(), max.z());
    v = Vector (0,max.y()-min.y(),0);
    break;
  case 2:
    pts[0] = min;
    pts[1] = Point (max.x(), min.y(), min.z());
    pts[2] = Point (max.x(), max.y(), min.z());
    pts[3] = Point (min.x(), max.y(), min.z());
    v = Vector (0,0,max.z()-min.z());
    break;
  }

  for (int i=0; i<4; i++)
  {
    if (sort_start)
    {
      pts[i] += v;
    }
  }

  glEnable(GL_TEXTURE_2D);
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
  glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,
            GL_MODULATE);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                  GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,
                  GL_LINEAR);

  // Make vector be scaled based on slice distance.
  switch (which)
  {
  case 0: // x
    v = v*sort_dir*1.0/nx;
    break;
  case 1: // y
    v = v*sort_dir*1.0/ny;
    break;
  case 2: // z
    v = v*sort_dir*1.0/nz;
    break;
  }

  // Get GL stuff set up.
  glColor4f(1,1,1,bright);

  glAlphaFunc(GL_GEQUAL,accum);  // This might be to large.
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  glEnable(GL_ALPHA_TEST);

  for (int outer=sort_start; outer != sort_end; outer += sort_dir)
  {
    switch (which)
    {
    case 0:     // x
      glTexImage2D(GL_TEXTURE_2D, 0, GL_INTENSITY, ny, nz, 0,
                   GL_LUMINANCE, GL_UNSIGNED_BYTE, &(Xmajor(outer,0,0)));
      break;
    case 1: // y
      glTexImage2D(GL_TEXTURE_2D, 0, GL_INTENSITY, nx, nz, 0,
                   GL_LUMINANCE, GL_UNSIGNED_BYTE, &(Ymajor(outer,0,0)));
      break;
    case 2: // y
      glTexImage2D(GL_TEXTURE_2D, 0, GL_INTENSITY, nx, ny, 0,
                   GL_LUMINANCE, GL_UNSIGNED_BYTE, &(Zmajor(outer,0,0)));
      break;
    }
        
    // now draw the quad for this texture.

    for (int j=0;j<4;j++)  // v has been rescaled based on dir
      pts[j] += v;      

    glBegin(GL_QUADS);
    glTexCoord2f(0,0);
    glVertex3f(pts[0].x(),pts[0].y(),pts[0].z());

    glTexCoord2f(0,1);
    glVertex3f(pts[1].x(),pts[1].y(),pts[1].z());

    glTexCoord2f(1,1);
    glVertex3f(pts[2].x(),pts[2].y(),pts[2].z());

    glTexCoord2f(1,0);
    glVertex3f(pts[3].x(),pts[3].y(),pts[3].z());
    glEnd();
  }

  glDisable(GL_BLEND);
  glDisable(GL_ALPHA_TEST);
  glDisable(GL_TEXTURE_2D);
  post_draw(di);
}

} // End namespace SCIRun

