/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2009 Scientific Computing and Imaging Institute,
   University of Utah.

   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/



/*
 *  TriStrip.cc: Triangle Strip object
 *
 *  Written by:
 *   Steven G. Parker & David Weinstein
 *   Department of Computer Science
 *   University of Utah
 *   April 1994
 *
 */

#include <Core/Util/Debug.h>

#include <Core/Geom/DrawInfoOpenGL.h>
#include <Core/Geom/GeomTriStrip.h>
#include <Core/Geom/GeomTri.h>

namespace SCIRun {

GeomTriStrip::GeomTriStrip()
{
  DEBUG_CONSTRUCTOR("GeomTriStrip")
}

GeomTriStrip::GeomTriStrip(const GeomTriStrip& copy)
: GeomVertexPrim(copy)
{
  DEBUG_CONSTRUCTOR("GeomTriStrip")
}

GeomTriStrip::~GeomTriStrip() 
{
  DEBUG_DESTRUCTOR("GeomTriStrip")
}

GeomObj* 
GeomTriStrip::clone()
{
  return new GeomTriStrip(*this);
}

void
GeomTriStrip::draw(DrawInfoOpenGL* di, Material* matl, double)
{
  if (!pre_draw(di, matl, 1)) return;
  if (verts.size() <= 2)
    return;
  di->polycount_+=verts.size()-2;
  if (di->currently_lit_)
  {
    glDisable(GL_NORMALIZE);
    switch(di->get_drawtype())
    {
    case DrawInfoOpenGL::WireFrame:
      {
        verts[0]->emit_all(di);
        verts[1]->emit_all(di);
        for (int i=2;i<verts.size();i++)
        {
          glBegin(GL_LINE_LOOP);
          verts[i-2]->emit_all(di);
          verts[i-1]->emit_all(di);
          verts[i]->emit_all(di);
          glEnd();
        }
      }
      break;
    case DrawInfoOpenGL::Flat:
    case DrawInfoOpenGL::Gouraud:
      {
        glBegin(GL_TRIANGLE_STRIP);
        for (int i=0;i<verts.size();i++)
        {
          verts[i]->emit_all(di);
        }
        glEnd();
      }
      break;
    }
    glEnable(GL_NORMALIZE);
  }
  else
  {
    switch(di->get_drawtype())
    {
    case DrawInfoOpenGL::WireFrame:
      {
        verts[0]->emit_material(di);
        verts[0]->emit_point(di);
        verts[1]->emit_material(di);
        verts[1]->emit_point(di);
        for (int i=2;i<verts.size();i++)
        {
          glBegin(GL_LINE_LOOP);
          verts[i-2]->emit_material(di);
          verts[i-2]->emit_point(di);
          verts[i-1]->emit_material(di);
          verts[i-1]->emit_point(di);
          verts[i]->emit_material(di);
          verts[i]->emit_point(di);
          glEnd();
        }
      }
      break;
    case DrawInfoOpenGL::Flat:
    case DrawInfoOpenGL::Gouraud:
      {
        glBegin(GL_TRIANGLE_STRIP);
        for (int i=0;i<verts.size();i++)
        {
          verts[i]->emit_material(di);
          verts[i]->emit_point(di);
        }
        glEnd();
      }
      break;
    }
  }
  post_draw(di);
}

int 
GeomTriStrip::size(void)
{
    return verts.size();
}

GeomTriStripList::GeomTriStripList()
:n_strips(0)
{
  DEBUG_CONSTRUCTOR("GeomTriStripList")
}

GeomTriStripList::~GeomTriStripList()
{
  // everything should be deleted
  DEBUG_DESTRUCTOR("GeomTriStripList")
}

GeomObj* 
GeomTriStripList::clone()
{
  return new GeomTriStripList(*this);
}

void 
GeomTriStripList::add(const Point& p)
{
  int last = pts.size();
  pts.grow(3);
  pts[last++] = float(p.x());
  pts[last++] = float(p.y());
  pts[last++] = float(p.z());
}


void 
GeomTriStripList::add(const Point& p, const Vector& v)
{
  int last = pts.size();
  pts.grow(3);

  pts[last++] = float(p.x());
  pts[last++] = float(p.y());
  pts[last++] = float(p.z());

  last = nrmls.size();
  nrmls.grow(3);
  nrmls[last++] = float(v.x());
  nrmls[last++] = float(v.y());
  nrmls[last++] = float(v.z());
}

void 
GeomTriStripList::end_strip(void)
{
  int last = pts.size();
  int last_s = strips.size();
  
  strips.grow(1);
  strips[last_s] = last;
}

int
GeomTriStripList::size(void)
{
  return nrmls.size()/3;
}

void
GeomTriStripList::draw(DrawInfoOpenGL* di, Material* matl, double)
{
  if (pts.size() == 0)
    return;
  if (!pre_draw(di, matl, 1)) return;

  di->polycount_ += size();
  if (di->currently_lit_)
  {
    glEnable(GL_NORMALIZE);

    switch(di->get_drawtype())
    {
    case DrawInfoOpenGL::WireFrame:
    case DrawInfoOpenGL::Flat:
    case DrawInfoOpenGL::Gouraud:
      {
        int nstrips = strips.size();
        int index=0;
        float *rpts = &pts[0];
        float *nrm = &nrmls[0];
        for (int ns = 0;ns < nstrips; ns++)
        {
          glBegin(GL_TRIANGLE_STRIP);
          glVertex3fv(rpts);
          rpts += 3;
          glVertex3fv(rpts);
          rpts += 3;
          index += 6;
          while (index < strips[ns])
          {
            glNormal3fv(nrm);
            nrm += 3;
            index += 3;
            glVertex3fv(rpts);
            rpts += 3;
          }
          glEnd();
        }
      }
      break;
    }
  }
  else
  {
    int nstrips = strips.size();
    int index=0;
    float *rpts = &pts[0];
    for (int ns = 0;ns < nstrips; ns++)
    {
      glBegin(GL_TRIANGLE_STRIP);
      glVertex3fv(rpts);
      rpts += 3;
      glVertex3fv(rpts);
      rpts += 3;
      index += 6;
      while (index < strips[ns])
      {
        index += 3;
        glVertex3fv(rpts);
        rpts += 3;
      }
      glEnd();
    }
  }
  post_draw(di);
}

Point 
GeomTriStripList::get_pm1(void)
{
  int npts = pts.size()-1;
  Point p(pts[npts-2],pts[npts-1],pts[npts]);
  return p;
}

Point
GeomTriStripList::get_pm2(void)
{
  int npts = pts.size()-1-3;
  Point p(pts[npts-2],pts[npts-1],pts[npts]);
  return p;
}

void 
GeomTriStripList::permute(int i0,int i1, int i2)
{
  float *ptr = &pts[pts.size()-1-8];
  float remap[9];

  remap[0] = ptr[i0*3 + 0];
  remap[1] = ptr[i0*3 + 1];
  remap[2] = ptr[i0*3 + 2];

  remap[3] = ptr[i1*3 + 0];
  remap[4] = ptr[i1*3 + 1];
  remap[5] = ptr[i1*3 + 2];

  remap[6] = ptr[i2*3 + 0];
  remap[7] = ptr[i2*3 + 1];
  remap[8] = ptr[i2*3 + 2];

  for(int i=0;i<9;i++) ptr[i] = remap[i];
}

void 
GeomTriStripList::get_bounds(BBox& box)
{
  for(int i=0;i<pts.size();i+=3)
    box.extend(Point(pts[i],pts[i+1],pts[i+2]));
}

int 
GeomTriStripList::num_since(void)
{
  int ssize = strips.size();
  int n_on = pts.size()/3;
  if (ssize) 
  {
    int lastp = strips[ssize-1];
    n_on -= lastp/3;
  }

  return n_on;
}

} // End namespace SCIRun

