/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2009 Scientific Computing and Imaging Institute,
   University of Utah.

   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/



/*
 *  GeomTube.cc: Tube object
 *
 *  Written by:
 *   Han-Wei Shen
 *   Department of Computer Science
 *   University of Utah
 *   Oct 1994
 *
 */

#include <Core/Util/Debug.h>
#include <Core/Geom/GeomTube.h>
#include <Core/Geom/GeomLine.h>
#include <Core/Geom/DrawInfoOpenGL.h>

#include <Core/Math/TrigTable.h>

namespace SCIRun {

GeomTube::GeomTube(int nu)
: nu(nu)
{
  DEBUG_CONSTRUCTOR("GeomTube")
}

GeomTube::GeomTube(const GeomTube& copy)
: GeomVertexPrim(copy), nu(copy.nu),
  directions(copy.directions), radii(copy.radii)
{
  DEBUG_CONSTRUCTOR("GeomTube")
}

GeomTube::~GeomTube()
{
  DEBUG_DESTRUCTOR("GeomTube")
}

GeomObj* GeomTube::clone() 
{
  return new GeomTube(*this); 
}

void  GeomTube::get_bounds(BBox& bb)
{
  for(int i=0;i<verts.size();i++)
  {
    bb.extend_disk(verts[i]->p, directions[i], radii[i]);
  }
}

// the function to extend the length of the tube geometry
void GeomTube::add(GeomVertex* vtx, double radius, const Vector& dir)
{
  GeomVertexPrim::add(vtx);  // Add the vertex - point and maybe color
  radii.add(radius);     // specify the radius of that point
  directions.add(dir.safe_normal());    // and the direction 
}

// Given a center point and its normal, compute those points on the 
// circle, this is a private member function, called  by 
// objdraw function
void GeomTube::make_circle(int which, Array1<Point>& circle_pts,
			   const SinCosTable& tab)
{
  Vector dir (directions[which]);
  double u = dir.x();
  double v = dir.y();
  double w = dir.z(); 
  double mat[6];
  if(w < -.999999)
  {
    mat[0]=-1;
    mat[1]=0;
    mat[2]=0;
    mat[3]=0;
    mat[4]=-1;
    mat[5]=0;
  } 
  else
  {
    double w1=1+w;
    mat[0]=v*v/w1 + w;
    mat[1]=-u*v/w1;
    mat[2]=-u;
    mat[3]=mat[1];
    mat[4]=u*u/w1 + w;
    mat[5]=-v;
  }

  circle_pts.remove_all();
  Point pt(verts[which]->p);
  double radius=radii[which];
  for (int i=0; i<=nu; i++) 
  { // temporarily set the number of grids as 20
    double cx = tab.sin(i)*radius;
    double cy = tab.cos(i)*radius;
    Point circle_pt(cx * mat[0] + cy*mat[3] + pt.x(),
			cx * mat[1] + cy*mat[4] + pt.y(),
			cx * mat[2] + cy*mat[5] + pt.z()); 
    circle_pts.add(circle_pt); 
  }
}


void
GeomTube::draw(DrawInfoOpenGL* di, Material* matl, double)
{
  if (!pre_draw(di, matl, 1)) return;
  
  di->polycount_+=(verts.size()-1)*2*20;
  Array1<Point> circle1;
  Array1<Point> circle2;
  Array1<Point>* p1=&circle1;
  Array1<Point>* p2=&circle2;
  SinCosTable tab(nu+1, 0, 2*M_PI);
  make_circle(0, *p1, tab);
  if (di->get_drawtype() == DrawInfoOpenGL::WireFrame)
  {
    glBegin(GL_LINE_LOOP);
    for (int j=0;j<p1->size();j++)
    {
      Point pt2((*p1)[j]);
      glVertex3d(pt2.x(), pt2.y(), pt2.z());
    }
    glEnd();
  }
  for (int i=0; i<verts.size()-1; i++)
  {
    make_circle(i+1, *p2, tab);
    Array1<Point>& pp1=*p1;
    Array1<Point>& pp2=*p2;
    switch(di->get_drawtype())
    {
    case DrawInfoOpenGL::WireFrame:
      {
        // Draw lines
        glBegin(GL_LINES);
        int j;
        for (j=0;j<nu;j++)
        {
          Point pt1(pp1[j]);
          glVertex3d(pt1.x(), pt1.y(), pt1.z());
          Point pt2(pp2[j]);
          glVertex3d(pt2.x(), pt2.y(), pt2.z());
        }
        glEnd();
        glBegin(GL_LINE_LOOP);
        for (j=0;j<nu;j++)
        {
          Point pt2(pp2[j]);
          glVertex3d(pt2.x(), pt2.y(), pt2.z());
        }
        glEnd();
      }
      break;
    case DrawInfoOpenGL::Flat:
    case DrawInfoOpenGL::Gouraud:
      {
        // draw triangle strips
        glBegin(GL_TRIANGLE_STRIP);
        Point cen1(verts[i]->p);
        Point cen2(verts[i+1]->p);
        for (int j=0;j<=nu;j++)
        {
          Point pt1(pp1[j]);
          Vector n1(pt1-cen1);
          verts[i]->emit_material(di);
          glNormal3d(n1.x(), n1.y(), n1.z());
          glVertex3d(pt1.x(), pt1.y(), pt1.z());

          Point pt2(pp2[j]);
          Vector n2(pt2-cen2);
          verts[i+1]->emit_material(di);
          glNormal3d(n2.x(), n2.y(), n2.z());
          glVertex3d(pt2.x(), pt2.y(), pt2.z());
        }
        glEnd();
      }
    }
    // Swith p1 and p2 pointers
    Array1<Point>* tmp=p1;
    p1=p2;
    p2=tmp;
  }
  post_draw(di);
}

} // End namespace SCIRun


