/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2009 Scientific Computing and Imaging Institute,
   University of Utah.

   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/

/*
 *  SimpleTextFileToMatrix_Plugin.cc
 *
 *  Written by:
 *   Jeroen Stinstra
 *   Department of Computer Science
 *   University of Utah
 *
 */
 

// This is a plugin is meant to read simple text files with pure data
// in ascii format into a SCIRun matrix.

#include <Core/ImportExport/Matrix/MatrixIEPlugin.h>
#include <Core/Datatypes/DenseMatrix.h>
#include <Core/Util/StringUtil.h>

#include <iostream>
#include <fstream>
#include <sstream>

namespace SCIRun {

MatrixHandle SimpleTextFileMatrix_reader(ProgressReporter *pr, const char *filename);
bool SimpleTextFileMatrix_writer(ProgressReporter *pr, MatrixHandle matrix, const char *filename);

MatrixHandle SimpleTextFileMatrix_reader(ProgressReporter *pr, const char *filename)
{
  MatrixHandle result;

  SCIRun::size_type ncols = 0;
  SCIRun::size_type nrows = 0;
  SCIRun::size_type line_ncols = 0;
  
  std::string line;
   
  // STAGE 1 - SCAN THE FILE TO DETERMINE THE DIMENSIONS OF THE MATRIX
  // AND CHECK THE FILE'S INTEGRITY.

  std::vector<double> values;

  {
    std::ifstream inputfile;
  
    try
    {
      inputfile.open(filename);
    }
    catch (...)
    {
      if (pr) pr->error("Could not open file: "+std::string(filename));
      return (result);
    }

    while( getline(inputfile,line,'\n'))
    {
			if (line.size() > 0)
			{
				// block out comments
				if ((line[0] == '#')||(line[0] == '%')) continue;
			}
		
      // replace comma's and tabs with white spaces
      for (size_t p = 0;p<line.size();p++)
      {
        if ((line[p] == '\t')||(line[p] == ',')||(line[p]=='"')) line[p] = ' ';
      }
      
      multiple_from_string(line,values);
      line_ncols = values.size();
      
      if (line_ncols > 0)
      {
        nrows++;
        if (ncols > 0)
        {
          if (ncols != line_ncols)
          {
            if (pr)  pr->error("Improper format of text file, not every line contains the same amount of numbers");
            return (result);
          }
        }
        else
        {
          ncols = line_ncols;
        }
      }
    }

    inputfile.close();
  }

  // STAGE 2 - NOW ACTUALLY READ AND STORE THE MATRIX

  {
    std::ifstream inputfile;

    result = dynamic_cast<Matrix *>(new DenseMatrix(nrows,ncols));
    if (result.get_rep() == 0)
    {
      if (pr) pr->error("Could not allocate matrix");
      return(result);
    }
    
    double* dataptr = result->get_data_pointer();
    int k = 0;

    try
    {
      inputfile.open(filename);
    }
    catch (...)
    {
      if (pr) pr->error("Could not open file: "+std::string(filename));
      return (result);
    }
    
    while( getline(inputfile,line,'\n'))
    {
			if (line.size() > 0)
			{
				// block out comments
				if ((line[0] == '#')||(line[0] == '%')) continue;
			}
					
      // replace comma's and tabs with white spaces
      for (size_t p = 0;p<line.size();p++)
      {
        if ((line[p] == '\t')||(line[p] == ',')||(line[p]=='"')) line[p] = ' ';
      }
      
      multiple_from_string(line,values);
      for (size_t j=0; j<values.size();j++) dataptr[k++] = values[j];
    }

    inputfile.close();
  }
  return(result);
}


bool SimpleTextFileMatrix_writer(ProgressReporter *pr, MatrixHandle matrix, const char *filename)
{

  std::ofstream outputfile;

  MatrixHandle temp = matrix->dense();
  matrix = temp;
  
  if (matrix.get_rep() == 0)
  {
    if (pr) pr->error("Empty matrix detected");
    return(false);
  }

  double* dataptr = matrix->get_data_pointer();
  if (dataptr == 0)
  {
    if (pr) pr->error("Empty matrix detected");
    return(false);
  }

  try
  {
    outputfile.open(filename);
  }
  catch (...)
  {
    if (pr) pr->error("Could not open file: "+std::string(filename));
    return (false);
  }
    
  size_t k = 0;
  for (int p=0; p<matrix->nrows(); p++)  
  {
    for (int q=0; q<matrix->ncols(); q++)  
    {
      outputfile << dataptr[k++] << " ";
    }
    outputfile << "\n";
  }  
  
  return (true);
}

static MatrixIEPlugin SimpleTextFileMatrix_plugin("SimpleTextFile","", "",SimpleTextFileMatrix_reader,SimpleTextFileMatrix_writer);

} // end namespace

