/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2009 Scientific Computing and Imaging Institute,
   University of Utah.

   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/


/*
 *  BuildEITMatrix.cc: solves EIT forward problems based on BEM.
 *
 *  Written by:
 *   Saeed Babaeizadeh - Northeastern University
 *   April 2006
 *
 */

#include <Packages/BioPSE/Core/Algorithms/NumApproximation/BuildBEMatrix.h>
#include <Dataflow/Network/Module.h>
#include <Dataflow/Network/Ports/MatrixPort.h>
#include <Dataflow/Network/Ports/FieldPort.h>

#include <Core/Datatypes/Matrix.h>
#include <Core/Datatypes/DenseMatrix.h>
#include <Core/Datatypes/ColumnMatrix.h>

#include <Core/Datatypes/Field.h>
#include <Core/Datatypes/Mesh.h>

#include <Core/Geometry/Vector.h>
#include <Core/Geometry/Point.h>
#include <Core/Geometry/BBox.h>
#include <Core/Util/StringUtil.h>

#include <Core/Math/MiscMath.h>

#include <algorithm>
#include <map>
#include <iostream>
#include <string>
#include <fstream>

#define epsilon 1e-12

namespace BioPSE {

using namespace SCIRun;

class BuildEITMatrix : public Module, public BuildBEMatrixBase
{
    MatrixHandle       hZoi_;
    typedef LockingHandle<DenseMatrix>     DenseMatrixHandle;

  private:

    bool ray_triangle_intersect(double &t,
              const Point &p,
              const Vector &v,
              const Point &p0,
              const Point &p1,
              const Point &p2) const;
    void compute_intersections(std::vector<std::pair<double, int> > &results,
                               const VMesh* mesh,
                               const Point &p, 
                               const Vector &v,
                               int marker) const;

    int compute_parent(const std::vector<VMesh*> &meshes, int index);

    bool compute_nesting(std::vector<int> &nesting,
                         const std::vector<VMesh*> &meshes);

    void build_Zoi( const std::vector<VMesh*>&,
                          std::vector<int>&,
                          std::vector<double>&,
                          MatrixHandle &);

  public:

    //! Constructor
    BuildEITMatrix(GuiContext *context);

    //! Destructor
    virtual ~BuildEITMatrix() {}
    virtual void execute();
};

DECLARE_MAKER(BuildEITMatrix)

// TODO: Modules need to be pure for the dataflow model and should not
// rely on globals

unsigned int first_time_run_EIT = 1;
std::vector<int> filed_generation_no_old_EIT, old_nesting_EIT;
std::vector<double> old_conductivities_EIT;

BuildEITMatrix::BuildEITMatrix(GuiContext *context):
  Module("BuildEITMatrix", context, Source, "Forward", "BioPSE")
{
}

void
BuildEITMatrix::execute()
{
  std::vector<FieldHandle> inputs;
  get_dynamic_input_handles("Surface",inputs,true);

  // Gather up the surfaces from the input ports.
  std::vector<FieldHandle> fields;
  std::vector<VMesh*> meshes;
  std::vector<double> conductivities;
  std::vector<int> filed_generation_no_new;
  std::string condStr; 
  double condVal;
  
  for (size_t j=0; j<inputs.size(); j++)
  {
    FieldHandle field = inputs[j];
    if (field.get_rep() == 0)
    {
      warning("Surface port '" + to_string(j) + "' contained no data.");
      continue;
    }


    VMesh *mesh = field->vmesh();
    if (!(mesh->is_trisurfmesh()))
    {
      error("Surface port '" + to_string(j) +
            "' does not contain a TriSurfField");
      return;
    }

    if (!field->get_property("Inside Conductivity", condStr))
    {
      error("The 'Inside Conductivity' of the Surface port '" + to_string(j) + "' was not set. It assumes to be zero!");
      condVal = 0;
    }
    else from_string(condStr,condVal);

    fields.push_back(field);
    meshes.push_back(mesh);
    conductivities.push_back(Abs(condVal));
    if(first_time_run_EIT)
    {
      filed_generation_no_old_EIT.push_back(-1);
      old_conductivities_EIT.push_back(-1);
      old_nesting_EIT.push_back(-1);
    }
    filed_generation_no_new.push_back(field->generation);
  }

  first_time_run_EIT = 0;

  // Compute the nesting tree for the input meshes.
  std::vector<int> nesting;
  if (!compute_nesting(nesting, meshes))
  {
    error("Unable to compute a valid nesting for this set of surfaces.");
  }

  // Check to see if the input fields are new
  int new_fields = 0, new_nesting = 0;
  double new_conductivities = 0;
  int no_of_fields =  nesting.size();
  for (int i=0; i < no_of_fields; i++)
  {
    new_fields += Abs( filed_generation_no_new[i] - filed_generation_no_old_EIT[i] );
    new_nesting += Abs( nesting[i] - old_nesting_EIT[i] );
    new_conductivities += Abs( conductivities[i] - old_conductivities_EIT[i] );
    filed_generation_no_old_EIT[i] = filed_generation_no_new[i];
    old_nesting_EIT[i] = nesting[i];
    old_conductivities_EIT[i] = conductivities[i];
  }

   
  if(new_fields>(no_of_fields+2) || new_nesting || new_conductivities )  // If the input fields are new
    build_Zoi(meshes, nesting, conductivities, hZoi_);
  else
    remark("Field inputs are old. Resending stored matrix.");

  // -- sending handles to cloned objects
  MatrixHandle hzoi(hZoi_->clone());
  send_output_handle("BEM Forward Matrix", hzoi);

  return;
}

// C++ized MollerTrumbore97 Ray Triangle intersection test.
#define EPSILON 1.0e-6
bool
BuildEITMatrix::ray_triangle_intersect(double &t,
				       const Point &point,
				       const Vector &dir,
				       const Point &p0,
				       const Point &p1,
				       const Point &p2) const
{
  // Find vectors for two edges sharing p0.
  const Vector edge1 = p1 - p0;
  const Vector edge2 = p2 - p0;

  // begin calculating determinant - also used to calculate U parameter.
  const Vector pvec = Cross(dir, edge2);

  // if determinant is near zero, ray lies in plane of triangle.
  const double det = Dot(edge1, pvec);
  if (det > -EPSILON && det < EPSILON)
  {
    return false;
  }
  const double inv_det = 1.0 / det;

  // Calculate distance from vert0 to ray origin.
  const Vector tvec = point - p0;

  // Calculate U parameter and test bounds.
  const double u = Dot(tvec, pvec) * inv_det;
  if (u < 0.0 || u > 1.0)
  {
    return false;
  }

  // Prepare to test V parameter.
  const Vector qvec = Cross(tvec, edge1);

  // Calculate V parameter and test bounds.
  const double v = Dot(dir, qvec) * inv_det;
  if (v < 0.0 || u + v > 1.0)
  {
    return false;
  }

  // Calculate t, ray intersects triangle.
  t = Dot(edge2, qvec) * inv_det;

  return true;
}


void
BuildEITMatrix::compute_intersections(std::vector<std::pair<double, int> > &results,
				      const VMesh* mesh,
				      const Point &p, 
              const Vector &v,
				      int marker) const
{
  VMesh::Elem::iterator itr, eitr;
  mesh->begin(itr);
  mesh->end(eitr);
  double t;
  while (itr != eitr)
  {
    VMesh::Node::array_type nodes;
    mesh->get_nodes(nodes, *itr);
    Point p0, p1, p2;
    mesh->get_center(p0, nodes[0]);
    mesh->get_center(p1, nodes[1]);
    mesh->get_center(p2, nodes[2]);
    if (ray_triangle_intersect(t, p, v, p0, p1, p2))
    {
      results.push_back(pair<double, int>(t, marker));
    }
    ++itr;
  }
}

static bool
pair_less(const pair<double, int> &a, const pair<double, int> &b)
{
  return a.first < b.first;
}


int
BuildEITMatrix::compute_parent(const std::vector<VMesh*> &meshes, int index)
{
  Point point;
  meshes[index]->get_center(point, VMesh::Node::index_type(0));
  Vector dir(1.0, 1.0, 1.0);
  std::vector<std::pair<double, int> > intersections;

  for (size_t i = 0; i < meshes.size(); i++)
  {
    compute_intersections(intersections, meshes[i], point, dir, i);
  }

  std::sort(intersections.begin(), intersections.end(), pair_less);

  std::vector<int> counts(meshes.size(), 0);
  for (size_t i = 0; i < intersections.size(); i++)
  {
    if (intersections[i].second == index)
    {
      // First odd count is parent.
      for (int j = i-1; j >= 0; j--)
      {
        if (counts[intersections[j].second] & 1)
        {
          return intersections[j].second;
        }
      }
      // No odd parent, is outside.
      return static_cast<int>(meshes.size());
    }
    counts[intersections[i].second]++;
  }

  // Indeterminant, we should intersect with ourselves.
  return (int)meshes.size();
}



bool
BuildEITMatrix::compute_nesting(std::vector<int> &nesting,
                                const std::vector<VMesh*>& meshes)
{
  nesting.resize(meshes.size());

  for (size_t i = 0; i < (unsigned int)meshes.size(); i++)
  {
    nesting[i] = compute_parent(meshes, i);
  }

  return true;
}


void BuildEITMatrix::build_Zoi(const std::vector<VMesh*> &meshes,
                               std::vector<int> &nesting,
                               std::vector<double> &conductivities,
                               MatrixHandle &hZoi_)
{
  std::vector<DenseMatrixHandle> PP;
  DenseMatrixHandle h_PP_;
  int i ,j;
  double in_cond, out_cond, op_cond;
  int no_of_fields = nesting.size();

  // PP Begin
  for (i=0; i<no_of_fields; i++)
  {
    for (j=0; j<no_of_fields; j++)
    {
      if (i==j)
      {
        in_cond = conductivities[i];
        if (nesting[i] == no_of_fields)  // The outermost surface is different from the others.
        {
          out_cond = 0;
          op_cond = in_cond;
        }
        else
        {
          out_cond = conductivities[nesting[i]];
          op_cond = out_cond;
        }
        BuildBEMatrixBase::make_auto_P(meshes[i], h_PP_, out_cond, in_cond, op_cond);
      }
      else
      {
        in_cond = conductivities[j];
        if (nesting[j] == no_of_fields) out_cond = 0;
        else out_cond = conductivities[nesting[j]];

        if (nesting[i] == no_of_fields) op_cond = conductivities[i];
        else op_cond = conductivities[nesting[i]];

        BuildBEMatrixBase::make_cross_P(meshes[i], meshes[j], h_PP_, in_cond, out_cond, op_cond);
      }
      PP.push_back(h_PP_);
    }
  }
  // PP End

  std::vector<DenseMatrixHandle> GG;
  DenseMatrixHandle h_GG_;

  // GG Begin
  std::vector<double>   avInn_;
  int hs = 0;
  for (i=0; i<(int)nesting.size(); i++) 
	if (nesting[i] == (int)nesting.size()) hs = i;
  int given_currents_surface_index = hs; // This is the surface whose currents are given
  BuildBEMatrixBase::pre_calc_tri_areas(meshes[given_currents_surface_index], avInn_);

  for (i=0; i<no_of_fields; i++)
  {
    if (i==given_currents_surface_index)
    {
      in_cond = conductivities[i];
      if (nesting[i] == no_of_fields)  // The Outermost surface is different from the others.
      {
        out_cond = 0;
        op_cond = in_cond;
      }
      else
      {
        out_cond = conductivities[nesting[i]];
        op_cond = out_cond;
      }
      BuildBEMatrixBase::make_auto_G(meshes[i], h_GG_, out_cond, in_cond, op_cond, avInn_);
    }
    else
    {
      in_cond = conductivities[given_currents_surface_index];
      if (nesting[given_currents_surface_index] == no_of_fields) out_cond = 0;
      else out_cond = conductivities[nesting[given_currents_surface_index]];

      if (nesting[i] == no_of_fields) op_cond = conductivities[i];
      else op_cond = conductivities[nesting[i]];

      BuildBEMatrixBase::make_cross_G(meshes[i], meshes[given_currents_surface_index], h_GG_, in_cond, out_cond, op_cond, avInn_);
    }
    GG.push_back(h_GG_);
  }

  if (no_of_fields == 1) // only one surface
  {
    PP[hs]->invert(); 
    DenseMatrixHandle hZbb_ = new DenseMatrix(GG[hs]->nrows(), GG[hs]->ncols()); 
    Mult(*(hZbb_.get_rep()), *(PP[hs].get_rep()), *(GG[hs].get_rep())); 
    hZbb_->scalar_multiply(-1/conductivities[hs]);
    hZoi_ = hZbb_.get_rep();
  }
  else if (no_of_fields == 2)  // two surfaces (one internal inhomogeneity)
  {
    int ms;
    (hs == 0) ? ms =1 : ms = 0;
    PP[ms*(no_of_fields+1)]->invert(); 
    DenseMatrixHandle hZbb_ = new DenseMatrix(GG[hs]->nrows(), GG[hs]->ncols()); 
    DenseMatrix tmp1(PP[hs*(no_of_fields-1)+1]->nrows(), PP[hs*(no_of_fields-1)+1]->ncols()); 
    Mult(tmp1, *PP[hs*(no_of_fields-1)+1].get_rep(), *PP[ms*(no_of_fields+1)].get_rep());  
    DenseMatrix tmp2(PP[hs*(no_of_fields+1)]->nrows(), PP[hs*(no_of_fields+1)]->ncols()); 
    Mult(tmp2, tmp1, *(PP[ms*(no_of_fields-1)+1].get_rep()));
    Add(1,*PP[hs*(no_of_fields+1)].get_rep() , -1, tmp2);  
    PP[hs*(no_of_fields+1)]->invert(); 
    Mult(tmp2, tmp1, *(GG[ms].get_rep()));
    Add(1,*GG[hs].get_rep() , -1, tmp2);  
    Mult(*(hZbb_.get_rep()), *(PP[hs*(no_of_fields+1)].get_rep()), *(GG[hs].get_rep()));
    hZbb_->scalar_multiply(-1/conductivities[hs]);
    hZoi_ = hZbb_.get_rep();
   
  } // end else   if (no_of_fields==2)

  return;
}

} // end namespace BioPSE
